//-------------------------------------------------------BEGINHEADER--
// This file is part of LoopSim (arXiv:1006.2144).
//
// Copyright 2009-2011 Mathieu Rubin, Gavin P. Salam and Sebastian Sapeta
//
// This software is provided as a snapshot of the ongoing LoopSim
// research project. As such, you may not redistribute it without the
// authors' permission.
//
// If you use LoopSim as part of your scientific work, you should
// discuss and agree with the LoopSim authors how best to acknowledge
// LoopSim in your work (e.g. whether through a reference, or through
// joint authorship with the LoopSim authors). 
//
// To help guide LoopSim's proper use in its current early stage of
// development, a condition of use of LoopSim is that any results that
// you obtain with it must be shown to the LoopSim authors before
// they are made public.
//
// $Id: LSAnalysisLHC.cc 1921 2013-06-30 11:47:48Z sapeta $
//-------------------------------------------------------ENDHEADER----

#include <fstream>
#include "LSAnalysisLHC.hh"
#include "fastjet/Selector.hh"

//-----------------------------------------------------------------------------
// constructor
LSAnalysisLHC::LSAnalysisLHC(double R, double ptmin) : _total_cs(0.0) {

  _jet_def = new JetDefinition (antikt_algorithm, R);

  // cuts inspired by ATLAS and CMS results on WZ and Wj
  _cuts["eta_l_max"]  = 2.5;
  _cuts["pt_l_min"]   = 20.0;
  _cuts["mTW_min"]    = 40.0;
  _cuts["pt_nu_min"]  = 25.0;
  _cuts["y_jet_max"]  = 4.4;
  _cuts["pt_jet_min"] = ptmin;
  _cuts["dRlj_min"]   = 0.5;

  _histograms["ptW"   ].declare(0.0,1500.0,50.0);
  _histograms["ptjet0"].declare(0.0,1500.0,50.0);
  _histograms["HTjet" ].declare(0.0,1500.0,50.0);
  _histograms["HTtot" ].declare(0.0,1500.0,50.0);
  _histograms["ptl"   ].declare(0.0,1500.0,50.0);
  _histograms["ptnu"  ].declare(0.0,1500.0,50.0);

}

//-----------------------------------------------------------------------------
void LSAnalysisLHC::analyse(const Event& ls_evt, const bool orig) {

  // first check if the event has W boson and at least one parton 
  // if it does not it means that the boson was looped by LoopSim 
  // therefore it will not appear in the final state and 
  // the event should be rejected 
  if (!_has_W_boson_and_parton(ls_evt)) return;

  // lepton cuts section, first we extract lepton and neutrino from  W
  // then we check if they pass the cuts defined in the constructor
 
  // get the lepton and neutrino from the W boson
  PseudoJet lepton_from_W, neutrino;
  PseudoJet product0 = ls_evt.Wbosons()[0].decay_products()[0];
  PseudoJet product1 = ls_evt.Wbosons()[0].decay_products()[1];
  if (abs(product0.user_index())==11 && abs(product1.user_index())==12) {
    lepton_from_W = product0;
    neutrino = product1;
  } else if(abs(product1.user_index())==11 && abs(product0.user_index())==12){
    lepton_from_W = product1;
    neutrino = product0;
  } else {
    _something_went_wrong("in the lepton from W analysis part");
  }

  // add the event's weight to the total cross section
  double weight = ls_evt.weight;
   _total_cs += weight;

  // mTW selection
  double phi_l  = lepton_from_W.phi();
  double phi_nu = neutrino.phi();
  double pt_l  = lepton_from_W.perp();
  double pt_nu = neutrino.perp();
  double mTW = sqrt(2.0*pt_l*pt_nu*(1.0-cos(phi_l-phi_nu)));
  if (mTW < _cuts["mTW_min"]) return;

  // lepton rapidity selection
  if (abs(lepton_from_W.eta()) > _cuts["eta_l_max"]) return;


  // cuts on pt of lepton and neutrino from W decay
  if (lepton_from_W.perp() <_cuts["pt_l_min"] ||
      neutrino.perp() < _cuts["pt_nu_min"]) return;


  // jet analysis section, compute jet related quantities
  double HTjets = 0.0, ptjet0 = 0.0;
  //for (unsigned i=0;i<ls_evt.partons().size();i++)
  //  all_visible.push_back(ls_evt.partons()[i]);
  ClusterSequence cs(ls_evt.partons(), *_jet_def);
  vector<PseudoJet>  jets = cs.inclusive_jets();
  Selector jet_selector = SelectorAbsRapMax(_cuts["y_jet_max"]) &&
                          SelectorPtMin(_cuts["pt_jet_min"]);
  vector<PseudoJet>  selected_jets = sorted_by_pt(jet_selector(jets));

  // lepton-jet separation
  vector<PseudoJet> tmp_jets;
  for (unsigned i=0;i<selected_jets.size();i++) {
    double dist = lepton_from_W.delta_R(selected_jets[i]);
    if (dist > _cuts["dRlj_min"]) 
      tmp_jets.push_back(selected_jets[i]);
  }
  selected_jets = tmp_jets;

  if (selected_jets.size() >= 1 )  {
    ptjet0 = selected_jets[0].perp();
    HTjets = HT(selected_jets);
  } else { return; }


  // if we got to this point it means the event passes all the cuts
  // the last step is the filling all histograms
  double ptW   = ls_evt.Wbosons()[0].perp();
  double ptl   = lepton_from_W.perp();
  double ptnu  = neutrino.perp();
  double HTtot = HTjets + ptl + ptnu;
  _histograms["ptW"   ].add_entry(ptW, weight);
  _histograms["ptjet0"].add_entry(ptjet0, weight);
  _histograms["HTjet" ].add_entry(HTjets, weight);
  _histograms["HTtot" ].add_entry(HTtot, weight);
  _histograms["ptl"   ].add_entry(ptl, weight);
  _histograms["ptnu"  ].add_entry(ptnu, weight);

}

//-----------------------------------------------------------------------------
bool LSAnalysisLHC::_has_W_boson_and_parton(const Event& ls_evt) {
  int nW = ls_evt.Wbosons().size();
  int np = ls_evt.partons().size();
  assert(nW <2 && np < 5); // can't remember why I set this...
  if (nW == 1 && np > 0) 
     return true;
  else 
    return false;
}

//-----------------------------------------------------------------------------
// add some extra info to the description that goes to output files
string LSAnalysisLHC::description() {
  stringstream sstr;
  sstr << LSAnalysisBase::description() << endl;
  sstr << "# total_cs = " <<  _total_cs << endl;
  return sstr.str();
}
