#include <iostream>
#include <cmath>
#include <cstring>
#include <cstdlib>
#include <fstream>
#include <sstream>
#include "mcfm.hh"
using namespace std;

// user card files options
bool writelhe   = false;         // lhef writing flag
string lhefname = "dummy.lhe";   // name of the lhe output file
int nrwgt       = 0;             // power of HT^nrwgt reweighting
double htmin    = -1.0;          // silly dummy values
double htmax    = -1.0;
bool purevirt   = false;         // virtual contribtuion only (without LO)

// keep track of iterations and allow for passing 
// this information between functions
int current_itno = 1;  

//------------------------------------------------------------------------------
// MCFM extern fortran rutines and common blocks
//------------------------------------------------------------------------------
void getptildejet(const int nd, double pjet[4][mxpart]);
double ht_particles(int nd);
double reweight(const int & nd);
double reweight(const double & ht);
extern "C" {

// name of the lhe file
// common/lhe/lhefile
extern struct {
  char lhefile[255];
} lhe_;

// function writing lhe file output
void lhewrite_(const double p[4][mxpart], //< particle vector
               const double & wt,         //< event weight
	       const double & facscale,   //< factorization scale
	       const int & itno,          //< current interation number
	       const int & nd             //< dipole number
	      );

// called at the end the run to properly finish the lhe file
void lhewrite_exit_();

//----------------------------------------------------------------------
bool userincludedipole_(const int & nd, 
                        const double * ppart , 
                        const bool & mcfm_result) {

  double ht = ht_particles(nd);

  // reweight the event
  reweight_.reweight = reweight(ht);
  
  // decide whether or not to include the event based on ht cuts
  if (ht < htmin || ht > htmax) return false;
  else return mcfm_result;

}


//----------------------------------------------------------------------
void userplotter_( const double * pjet, 
                   const double & wt,
                   const double & wt2,
                   const int    & nd) {

  static bool first = true;
  
  string tag; // tag: "book" to initialize histogram, "plot" to fill
  if (first) {
    first = false;
    tag = "book";
    // set the name of the output lhe file
    char lhefname_array[255];
    sprintf(lhefname_array, "%s", lhefname.c_str());
    strcpy(lhe_.lhefile, lhefname_array);
  } else {
    tag = "plot";
  }

  // write lhe output
  if (writelhe && tag == "plot") {
    double ppart[4][mxpart];
    getptildejet(nd,ppart);
    if (!strcmp(part_.part, "lord") || !strcmp(part_.part, "virt")) {
      lhewrite_(ppart,wt,facscale_.facscale,current_itno, nd);
    } else if (!strcmp(part_.part, "real")) {
      lhewrite_(ppart,wt,dipolescale_.dipscale[nd],current_itno, nd);
    }
  }

}


//----------------------------------------------------------------------
void userwriteinfo_(const int & unitno,
                    const char * comment_chars, // an array of size 2
                    const double & xsec,
                    const double & xsec_err,
                    const int & itno) {
  mcfmfwrite(unitno,comment_chars, "mcfmcc program comments");
  
  {ostringstream ostr;
  ostr << "[writelhe] " << writelhe ;
  mcfmfwrite(unitno,comment_chars, ostr.str());}

  {ostringstream ostr;
  ostr << "[lhefname] " << lhefname;
  mcfmfwrite(unitno,comment_chars, ostr.str());}

  {ostringstream ostr;
  ostr << "[nrwgt] " << nrwgt;
  mcfmfwrite(unitno,comment_chars, ostr.str());}

  {ostringstream ostr;
  ostr  << "[purevirt] " << useropt_.purevirt;
  mcfmfwrite(unitno,comment_chars, ostr.str());}

}
 
//---------------------------------------------------------------------------
// Function called by histofin at the end of each iteration and at the end of
// the run of MCFM
void userhistofin_(const double & xsec,
                   const double & xsec_err,
		   const int & itno, 
		   const int & itmx) {
  // increment the iteration counter
  current_itno++;
  // itno goes from 1 to itmx but the last call of histofin 
  // in mcfm_exit() is done with itno = 0	    
  if (itno == 0) {
    lhewrite_exit_();
  } 
}

} // extern "C"

//------------------------------------------------------------------------------
// C++ custom rutines
//------------------------------------------------------------------------------
/// return the scalar sum of the pt's of all particles
double ht_particles(int nd) {
  double ht = 0.0;
  for (int i = 2; i < 2+npart_.npart; i++) {
    ht += sqrt(pow(ptildes_.ptilde[0][i][nd],2) +
               pow(ptildes_.ptilde[1][i][nd],2));
  }
  return ht;
}

//----------------------------------------------------------------------
/// read extra parameters from the file, assuming a notation of the form
/// [x value ...param...]
template<class T>
T read_param_from_file (const string & inputfile, const string & param_name) {
  ifstream fstr(inputfile.c_str());
  string line;
  while (getline(fstr,line)) {
    cout << line << endl;
    if (line.find(param_name) != string::npos) {
      istringstream istr(line);
      string dummy; 
      T result;
      istr >> dummy >> result;
      if (dummy != "[x") continue; // if the initial tag is not "[x" ignore this line
      return result;
    }
  }
  cerr << "Did not find a parameter " << param_name << endl;
  exit(EXIT_FAILURE);
} 

//----------------------------------------------------------------------------
/// a C++ alternative to MCFM's f77 determinefilenames_, so that one
/// can determine filenames also when the main program is in C++.
///
/// This roughly follows MCFM's pattern of command-line handling. 
/// Note: you MUST resize inputfile and workdir to 72 characters before
/// passing them to MCFM's mcfmmain(inputfile, workdir)
void determine_filenames(int argc, char ** argv,
                         std::string & inputfile, std::string & workdir) {
  if (argc == 1) {
    // no arguments: assume we're using input.DAT
    inputfile = "input.DAT";
    workdir = "";
  } else if (argc >= 2) {
    // 1 argument: direct link to the filename (MCFM would also allow
    // to be just the directory - we don't do that here; but we do
    // allow the file to be in a subdirectory, which is then used as
    // the working directory.)
    inputfile = argv[1];
    std::size_t slash_pos = inputfile.find_last_of('/');
    if (slash_pos != std::string::npos) {
      workdir = inputfile.substr(0, slash_pos);
    } else {
      workdir = "";
    }

  // the following lines give default MCFM behaviour: workdir filename
  // If you want this behaviour, replace argc>=2 above with argc==2
  //---------------------------------------------------
  // } else if (argc == 3) {
  //   // 2 arguments: working directory and filename within that directory
  //   workdir = argv[1];
  //   inputfile = argv[2];
  // 
  //   // remove trailing slash (as wanted by MCFM)
  //   if (workdir.length() > 1 &&
  //       workdir[workdir.length()-1] == '/') {
  //     workdir = workdir.substr(0,workdir.length());
  //   }
  //   inputfile = workdir+"/"+inputfile;
  }

  // here we do our own behaviour: 3rd argument indicates working directory 
  if (argc >= 3) {
    workdir = argv[2];
  }
}


//----------------------------------------------------------------------------
// This functions does in principle the same as the MCFM fucntion getptildejet_ 
// but the latter misbehaves here loosing some particles. Why?
void getptildejet(const int nd,
                 double pjet[4][mxpart]) {

  for (int i=0;i<4;i++)
    for (int j=0;j<mxpart;j++)
      pjet[i][j] = ptildes_.ptilde[i][j][nd];

}

// Reweight functions
double reweight(const int & nd) {

  return pow(ht_particles(nd),nrwgt);
}

double reweight(const double & ht) {

  return pow(ht,nrwgt);
}



//----------------------------------------------------------------------
int main(int argc, char ** argv) {
  string inputfile, workdir;
  determine_filenames(argc, argv, inputfile, workdir);
  
  cout << workdir << "|" << endl;
  cout << inputfile  << "|" << endl;

  writelhe = read_param_from_file<bool>(inputfile, "writelhe");
  lhefname = read_param_from_file<string>(inputfile, "lhefname");
  nrwgt    = read_param_from_file<int>(inputfile, "nrwgt");
  htmin    = read_param_from_file<double>(inputfile, "htmin");
  htmax    = read_param_from_file<double>(inputfile, "htmax");
  purevirt = read_param_from_file<bool>(inputfile, "purevirt");
  useropt_.purevirt  = purevirt;
  
  workdir.resize(72,' ');
  inputfile.resize(72,' ');
  
  mcfmmain_(inputfile.c_str(), workdir.c_str());
  
}

