//-------------------------------------------------------BEGINHEADER--
// This file is part of LoopSim (arXiv:1006.2144).
//
// Copyright 2009-2011 Mathieu Rubin, Gavin P. Salam and Sebastian Sapeta
//
// This software is provided as a snapshot of the ongoing LoopSim
// research project. As such, you may not redistribute it without the
// authors' permission.
//
// If you use LoopSim as part of your scientific work, you should
// discuss and agree with the LoopSim authors how best to acknowledge
// LoopSim in your work (e.g. whether through a reference, or through
// joint authorship with the LoopSim authors). 
//
// To help guide LoopSim's proper use in its current early stage of
// development, a condition of use of LoopSim is that any results that
// you obtain with it must be shown to the LoopSim authors before
// they are made public.
//
// $Id: LSAnalysisBase.cc 1999 2013-08-22 14:21:20Z sapeta $
//-------------------------------------------------------ENDHEADER----

#include "LSAnalysisBase.hh"
#include <fstream>

int LSAnalysisBase::iev_orig = -1;

//-----------------------------------------------------------------------------
void LSAnalysisBase::write(int iev, int iit, string header, 
                                 string outfilename) {
  
  cout << "Event " << iev << endl;
  ofstream ostr(outfilename.c_str());
  ostr << header;
  ostr << description();
  ostr << "# nev = " << iev << endl;
  ostr << "# Finished iteration " << iit << endl << endl;
  ostr.precision(10);

  // 1d histograms

  // Normalization
  // Events in MCFM are normalized such that summing up weights from
  // all events in a single iteration gives total cross section in [fb] 
  // and distributions in [fb/(bin size in GeV)]. To get it in
  // [fb/GeV] we divide by the bin size.  To be able to compare
  // intermediate results we need to divide the histograms by the
  // number of already finished iterations.

  for(map<string,SimpleHist>::const_iterator his_iter =_histograms.begin();
                his_iter != _histograms.end(); his_iter++) {
    ostr << "# " << (*his_iter).first << endl;
    double norm  = 1.0/(*his_iter).second.binsize()/((double) iit);
    output((*his_iter).second, &ostr, norm);
    ostr << "# total_weight " 
         << (*his_iter).second.total_weight()/((double) iit) << endl;
    ostr << endl << endl;
  }

  // 2d histograms

  for(map<string,SimpleHist2D>::const_iterator his_iter =_histograms2d.begin();
                his_iter != _histograms2d.end(); his_iter++) {
    ostr << "# " << (*his_iter).first << endl;
    double norm  = 
           1.0/(*his_iter).second.u_binsize()/(*his_iter).second.v_binsize()
           /((double) iit);
    output((*his_iter).second, &ostr, norm);
    ostr << "# total_weight " 
         << (*his_iter).second.total_weight()/((double) iit) << endl;
    ostr << endl << endl;
  }

  // 1d ordered histograms

  for(PresInsOrderMap<string,SimpleHist>::const_iterator his_iter =
      _ordered_histograms.begin(); his_iter != _ordered_histograms.end(); 
      his_iter++) {
    ostr << "# " << (*his_iter).first << endl;
    double norm  = 1.0/(*his_iter).second.binsize()/((double) iit);
    output((*his_iter).second, &ostr, norm);
    ostr << "# total_weight " 
         << (*his_iter).second.total_weight()/((double) iit) << endl;
    ostr << endl << endl;
  }
}

//-----------------------------------------------------------------------------
// get description including all cuts, jet algorithm
string LSAnalysisBase::description() {

  stringstream sstr;
  if (_jet_def !=0) {
    sstr << "# " << _jet_def->description() << endl;
  }
  for(map<string,double>::const_iterator cuts_iter =_cuts.begin();
                cuts_iter != _cuts.end(); cuts_iter++) {
    sstr << "# " << (*cuts_iter).first << " = " << (*cuts_iter).second << endl;
  }
  return sstr.str();
}

//-----------------------------------------------------------------------------
double LSAnalysisBase::HT(const vector<PseudoJet>& jets) {
  double ht = 0;
  for (unsigned i = 0; i < jets.size(); i++) {
    ht += jets[i].perp();
  }
  return ht;
}

//-----------------------------------------------------------------------------
// cancel all cuts by setting the mins to 0 and the maxes to infinity
// any run with cuts = 0 and makels = 0 must recover the original cross section
// from the lhe file (which is the sum of weights of all events in that file)
void LSAnalysisBase::_set_no_cuts() {

  for(map<string,double>::iterator cuts_iter =_cuts.begin();
  cuts_iter != _cuts.end(); cuts_iter++) {
    if ((*cuts_iter).first.find("min") != string::npos) {
      (*cuts_iter).second  = 0;
    } else if ((*cuts_iter).first.find("max") != string::npos) {
      (*cuts_iter).second  = 1e10;
    } else {
      _something_went_wrong ("in _set_no_cuts, no min or max in the name cut "
                            +(*cuts_iter).first);
    }
  }
}


//-----------------------------------------------------------------------------
void LSAnalysisBase::_something_went_wrong (string where) {
  cerr << "ERROR: LSAnalysisBase: "
       << "Something went wrong in the " << where <<  endl;
  exit(-1);
}
