//-------------------------------------------------------BEGINHEADER--
// This file is part of LoopSim (arXiv:1006.2144).
//
// Copyright 2009-2011 Mathieu Rubin, Gavin P. Salam and Sebastian Sapeta
//
// This software is provided as a snapshot of the ongoing LoopSim
// research project. As such, you may not redistribute it without the
// authors' permission.
//
// If you use LoopSim as part of your scientific work, you should
// discuss and agree with the LoopSim authors how best to acknowledge
// LoopSim in your work (e.g. whether through a reference, or through
// joint authorship with the LoopSim authors). 
//
// To help guide LoopSim's proper use in its current early stage of
// development, a condition of use of LoopSim is that any results that
// you obtain with it must be shown to the LoopSim authors before
// they are made public.
//
// $Id: LoopSimDriver.cc 1918 2013-06-29 14:07:12Z sapeta $
//-------------------------------------------------------ENDHEADER----

#include "LoopSimDriver.hh"

//-----------------------------------------------------------------------------
// constructor
// extract info from card file
//
// $ ./analysis input.dat < events.lhe
LoopSimDriver::LoopSimDriver(int argc, char ** argv, InputDat inputdat) :
                                              _event_source(NOSOURCE),_dR(0.5) {

  // extract information from command line
  string input_fname;
  // LHE files
  if (argc == 2) {
    _event_source = LHE;
  } else {cerr << "LoopSimDriver::ERROR: too few or tooo many arguments! " 
               << endl; exit(1);}

  // extract information from the card file 
  // and set options mandatory for LoopSimDriver
  _inputdat = new InputDat(inputdat);
  _set_mandatory_options();
  _check_event_source();

}
//-----------------------------------------------------------------------------
// old constructor; you are discouraged to use it
// $ ./analysis
// $ ./analysis input.dat < events.lhe
// $ ./analysis input.dat ntuples.root
LoopSimDriver::LoopSimDriver(int argc, char ** argv) :
                                              _event_source(NOSOURCE),_dR(0.5) {

  // extract information from command line
  string input_fname;
  if (argc==1) { 
    input_fname = "input.dat";
    _event_source = LHE;
  } else if (argc == 2) { 
    input_fname = string(argv[1]);
    _event_source = LHE;
  #ifdef NTUPLEROOT
  } else if (argc == 3) {
    input_fname = string(argv[1]);
    _ntuple_name = string(argv[2]);
    if (!_file_exists(_ntuple_name)) {
      cerr << "LoopSimDriver::ERROR: ntuple file does not exist! " 
           << endl; exit(1);}
    _event_source = NTUPLE;
  #endif // NTUPLEROOT
  } else {cerr << "LoopSimDriver::ERROR: too many arguments! " 
               << endl; exit(1);}

  // extract information from the card file 
  // and set options mandatory for LoopSimDriver
  _inputdat = new InputDat(input_fname);
  _set_mandatory_options();
  // set additional parameters
  _do_cuts  = _inputdat->value<bool>("cuts");
  _jet_veto = _inputdat->value<bool>("jetveto");
  _jet_veto2nd = _inputdat->value<bool>("jetveto2nd");
  _scale_factor = _inputdat->value<double>("scalefactor");

  _check_event_source();

}



//-----------------------------------------------------------------------------
void LoopSimDriver::_set_mandatory_options() {

  // extract information from line card file
  cout <<  _inputdat->description();
  _outfilename = _inputdat->value<string>("outfilename");
  _nborn = _inputdat->value<int>("nborn");
  _verbose = _inputdat->value<int>("verbose");
  _spread_virtuals = _inputdat->value<bool>("sv");
  _RLS = _inputdat->value<double>("RLS");
  _make_loopsim_analysis = _inputdat->value<bool>("makels");

  // find position of ".dat" in base output file name
  _datpos = _outfilename.find(".dat");

}

//-----------------------------------------------------------------------------
void LoopSimDriver::_check_event_source() {
  if (_event_source == NOSOURCE) {
     cerr << "ERROR: No event source!" << endl;
     exit(-1);
  }
}

//-----------------------------------------------------------------------------
// destructor
LoopSimDriver::~LoopSimDriver() {
  delete _inputdat; _inputdat = 0;
  for(map<string,LSAnalysisBase*>::iterator it =
      _lsanalysesRmin.begin(); it != _lsanalysesRmin.end(); it++) {
      delete (*it).second; (*it).second = 0;
  }
  for(map<string,LSAnalysisBase*>::iterator it =
      _lsanalysesRmid.begin(); it != _lsanalysesRmid.end(); it++) {
      delete (*it).second; (*it).second = 0;
  }
  for(map<string,LSAnalysisBase*>::iterator it =
      _lsanalysesRmax.begin(); it != _lsanalysesRmax.end(); it++) {
      delete (*it).second; (*it).second = 0;
  }
}

//-----------------------------------------------------------------------------
// functions for adding analyses to Rmin, Rmax or Rmid sets
void LoopSimDriver::add_analysisRmin(LSAnalysisBase* lsa, const string& name) {
  // analysis class
  _lsanalysesRmin[name.c_str()] = lsa;
  // output file name
  string tmp = _outfilename;
  tmp.insert(_datpos,name+string("Rmin"));
  _outfilenameRmin[name.c_str()] = tmp;
}

void LoopSimDriver::add_analysisRmax(LSAnalysisBase* lsa, const string& name) {
  // analysis class
  _lsanalysesRmax[name.c_str()] = lsa;
  // output file name
  string tmp = _outfilename;
  tmp.insert(_datpos,name+string("Rmax"));
  _outfilenameRmax[name.c_str()] = tmp;
}

void LoopSimDriver::add_analysisRmid(LSAnalysisBase* lsa, const string& name) {
  // analysis class
  _lsanalysesRmid[name.c_str()] = lsa;
  // output file name
  string tmp = _outfilename;
  tmp.insert(_datpos,name+string("Rmid"));
  _outfilenameRmid[name.c_str()] = tmp;
}


//-----------------------------------------------------------------------------
// main function doing the looping
void LoopSimDriver::run() {

  // read events in LHE format from stdin
  if (_event_source == LHE) {
      string line;
      ExtractLHEvent record;
      while(cin) {
        getline(cin, line);
        record.readline(line);
        if (!process_next_event(record)) break;
      }
   // read events from ntuple
   #ifdef NTUPLEROOT
   } else if (_event_source == NTUPLE) {
      ExtractNtupleEvent record(_scale_factor);
      record.add_file(_ntuple_name);
      while (record.has_next_event()) {
        if (!process_next_event(record)) break;
      }
   #endif //NTUPLEROOT
   } else if (_event_source == NOSOURCE) {
      cerr << "ERROR: No event source!" << endl;
      exit(-1);
  }
}

//-----------------------------------------------------------------------------
bool LoopSimDriver::process_next_event(ExtractEventBase& record) {
  static double total_cs_nocuts = 0.0;
  //----------------------------
  // read the new original event
  if(Event* this_evt = record.next_event()) {
    if (record.iev()==0)
     _generator = record.generator();
    total_cs_nocuts += this_evt->weight;

    // get order and nloops corresponding to that event
    int order=-1, nloops=-1; 
    record.current_event_extra_info().get_order_and_nloops(order, nloops);

    // set static class variable iev_orig in LSAnalysisBase
    map<string,LSAnalysisBase*>::const_iterator it = _lsanalysesRmid.begin();
    (*it).second->iev_orig = record.iev();

    //instantiate LoopSim
    LoopSim lsRmin(order,nloops,*this_evt,_RLS-_dR,_nborn,_spread_virtuals);
    LoopSim lsRmid(order,nloops,*this_evt,_RLS    ,_nborn,_spread_virtuals);
    LoopSim lsRmax(order,nloops,*this_evt,_RLS+_dR,_nborn,_spread_virtuals);
    
    // ----------------------------------------------------------
    // analyse LoopSim events three times for R-_dR, R and R+_dR
    while (lsRmin.there_is_a_next_event()) {
      const Event& ls_evt = lsRmin.extract_next_event();
      for(map<string,LSAnalysisBase*>::const_iterator it =
          _lsanalysesRmin.begin(); it != _lsanalysesRmin.end(); it++) {
          (*it).second->analyse(ls_evt,false);
      }
      if(!_make_loopsim_analysis) break;
    }
    int count = 0; 
    while (lsRmid.there_is_a_next_event()) {
      const Event& ls_evt = lsRmid.extract_next_event();
      if(_verbose) print_event(ls_evt,true,true);
      for(map<string,LSAnalysisBase*>::const_iterator it =
          _lsanalysesRmid.begin(); it != _lsanalysesRmid.end(); it++) {
          (*it).second->analyse(ls_evt,!(bool(count)));
      }
      if(!_make_loopsim_analysis) break;
      count++;
    }
    while (lsRmax.there_is_a_next_event()) {
      const Event& ls_evt = lsRmax.extract_next_event();
      for(map<string,LSAnalysisBase*>::const_iterator it =
          _lsanalysesRmax.begin(); it != _lsanalysesRmax.end(); it++) {
          (*it).second->analyse(ls_evt,false);
      }
      if(!_make_loopsim_analysis) break;
    }

    // Print the number of finished events and write the intermediate
    // results if possible.
    // MCFM: periodic_output is true only when we start new iteration 
    // and this is when we save results from all previous iterations
    // VBFNLO: the results are written down only in the last iteration;
    // it is important that the number of iterations in the extra line
    // of lhe file produced by VBFNLO is always = 1
    if ((_generator == "MCFM") && record.periodic_output()) {
        for(map<string,LSAnalysisBase*>::const_iterator it =
            _lsanalysesRmin.begin(); it != _lsanalysesRmin.end(); it++) {
            (*it).second->write(record.iev(),record.it()-1, 
	    _inputdat->description(), _outfilenameRmin[(*it).first]);
        }
        for(map<string,LSAnalysisBase*>::const_iterator it =
            _lsanalysesRmid.begin(); it != _lsanalysesRmid.end(); it++) {
            (*it).second->write(record.iev(),record.it()-1,
	    _inputdat->description(), _outfilenameRmid[(*it).first]);
        }
        for(map<string,LSAnalysisBase*>::const_iterator it =
            _lsanalysesRmax.begin(); it != _lsanalysesRmax.end(); it++) {
            (*it).second->write(record.iev(),record.it()-1,
	    _inputdat->description(), _outfilenameRmax[(*it).first]);
        }
    } else if ((_generator == "VBFNLO" || _generator == "NTUPLE" ) &&
               (record.iev()%10000 == 0)) {
        cout << "Event " << record.iev() << endl;
    }
  }

  // final writing after all events have been analysed
  // SS: at least in the case of NTUPLE it seems it would be more correct to
  // pass record.iev()+1 below since iev() starts from 0 and ends at
  // total_number_of_events -1; one should check if this applies to MCFM as well
  // and do the corresponding modifications
  if(record.end_of_file()) {
    for(map<string,LSAnalysisBase*>::const_iterator it =
        _lsanalysesRmin.begin(); it != _lsanalysesRmin.end(); it++) {
        (*it).second->write(record.iev(),record.it(),_inputdat->description(),
                            _outfilenameRmin[(*it).first]);
    }
    for(map<string,LSAnalysisBase*>::const_iterator it =
        _lsanalysesRmid.begin(); it != _lsanalysesRmid.end(); it++) {
        (*it).second->write(record.iev(),record.it(),_inputdat->description(),
                            _outfilenameRmid[(*it).first]);
    }
    for(map<string,LSAnalysisBase*>::const_iterator it =
        _lsanalysesRmax.begin(); it != _lsanalysesRmax.end(); it++) {
        (*it).second->write(record.iev(),record.it(),_inputdat->description(),
                            _outfilenameRmax[(*it).first]);
    }
    cout << "total_cs for the original process with no cuts = " 
         << total_cs_nocuts << endl;
    return false;
  }

  return true;
}

//-----------------------------------------------------------------------------
string LoopSimDriver::event_source() const {
  
  string val;
  switch(_event_source) { 
    case LHE:
    val = "LHE";
    break;
    case NTUPLE:
    val = "NTUPLE";
    break;
    case NOSOURCE:
    val = "NOSOURCE";
    break;
  }
  return val;
}

//-----------------------------------------------------------------------------
bool LoopSimDriver::_file_exists(string _file_name) const {
  fstream filestr;
  filestr.open(_file_name.c_str(),fstream::in);
  bool status = filestr.is_open();
  filestr.close();
  return status;
}
