//-------------------------------------------------------BEGINHEADER--
// This file is part of LoopSim (arXiv:1006.2144).
//
// Copyright 2009-2011 Mathieu Rubin, Gavin P. Salam and Sebastian Sapeta
//
// This software is provided as a snapshot of the ongoing LoopSim
// research project. As such, you may not redistribute it without the
// authors' permission.
//
// If you use LoopSim as part of your scientific work, you should
// discuss and agree with the LoopSim authors how best to acknowledge
// LoopSim in your work (e.g. whether through a reference, or through
// joint authorship with the LoopSim authors). 
//
// To help guide LoopSim's proper use in its current early stage of
// development, a condition of use of LoopSim is that any results that
// you obtain with it must be shown to the LoopSim authors before
// they are made public.
// 
// $Id: PresInsOrderMap.hh 1999 2013-08-22 14:21:20Z sapeta $
//-------------------------------------------------------ENDHEADER----


#ifndef __PRESINSORDERMAP_HH__
#define __PRESINSORDERMAP_HH__
#include <iostream>      
#include <string>
#include <map>
#include <vector>

using namespace std;
template <class keyType, class valType> class PresInsOrderMapIter;

/// Class providing a map which keeps the order of insertion.
///
/// The trick is to have two containers, a standard map and a vector. The
/// map keeps the pairs (key, val), as usual, and the vector keeps the keys 
/// in the order they were inserted.
///
/// Then, the iterator of PresInsOrderMap object, PresInsOrderMapIter, runs over
/// the vector entries rather than the map entries and gets the values from the
/// map using the key stored in vector. The interface to this class is identical
/// with that of the STL map.
template <class keyType, class valType> class PresInsOrderMap  {
  public:

  // declaration of associated iterator class
  friend class PresInsOrderMapIter<keyType,valType>;
  typedef PresInsOrderMapIter<keyType,valType>  const_iterator;

  /// element access operator
  // if key does not exist, add it to an internal vector and return reference 
  // to the key's value maped in the private STL map object
  valType& operator [] (const keyType& key) {
    if(_stlmap.find(key) == _stlmap.end()) {
      _ordervec.push_back(key);
    }
    return _stlmap[key];
  }

  // begin and end functions
  const_iterator begin() { 
     return PresInsOrderMapIter<keyType, valType>(_ordervec.begin(),&_stlmap); 
  }
  const_iterator end() { 
     return PresInsOrderMapIter<keyType, valType>(_ordervec.end(),&_stlmap); 
  }

  private:
  // STL map: key -> val; the main container
  map<keyType,valType> _stlmap; 
  // STL vector: int -> key; keeps the information about insertion order
  vector<keyType> _ordervec;    

};

/// Iterator of the PresInsOrderMap class.
///
/// Runs over vector entries rather than map entries. The value from the map
/// corresponding to a given key is determined at the level of dereferencing.
template <class keyType, class valType> 
class PresInsOrderMapIter : public iterator<forward_iterator_tag, valType> {

  public:
  friend class PresInstOrderMap;
  typedef typename vector<keyType>::iterator vec_iterator;

  PresInsOrderMapIter() {}
  PresInsOrderMapIter(vec_iterator vit, map<keyType,valType>  *stlmap_ptr) : 
                      _vit(vit), _stlmap_ptr(stlmap_ptr) {}

  /// dereference operator
  pair<keyType,valType> operator*() const {
    return pair<keyType,valType> (*_vit, _stlmap_ptr->operator[](*_vit));
  }

  /// prefix increment operator
  const PresInsOrderMapIter<keyType, valType>& operator++() {
    ++_vit;
    return *this;
  }

  /// postfix increment operator
  const PresInsOrderMapIter<keyType, valType> operator++(int) {
    PresInsOrderMapIter<keyType, valType> result(_vit, _stlmap_ptr);
    ++(*this);
    return result;
  }

  bool operator!=(const PresInsOrderMapIter<keyType, valType>& other) const {
    return this->_vit != other._vit;
  }

  private:

  // value of vector iterator specifies the position 
  // of PresInstOrderMap iterator
  vec_iterator _vit;
  map<keyType,valType> *_stlmap_ptr;

};


#endif // __PRESINSORDERMAP_H__
